import os
import unittest
import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.testing_utils import assert_cases_equal, assert_timeseries_near_equal


def create_linear_time_series(n, t_begin, t_end, x_begin, x_end):
    """
    Simple little function to generate a time series to be used for the tests in this file
    """
    slope = (x_end - x_begin) / (t_end - t_begin)
    line = lambda t: slope * (t - t_begin) + x_begin

    t = np.linspace(t_begin, t_end, n).reshape(n, 1)
    x = line(t)
    return t, x


class TestAssertTimeseriesNearEqual(unittest.TestCase):

    def test_assert_different_shape(self):
        t_ref = np.linspace(0, 100, 50)
        t_check = np.linspace(0, 100, 60)

        x_ref = np.atleast_2d(np.sin(t_ref)).T
        x_check = np.atleast_3d(np.sin(t_check)).T

        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-03)
        expected = "The shape of the variable in the two timeseries is not equal x_ref is (1," \
                   ")  x_check is (60, 1)"
        self.assertEqual(expected, str(e.exception))

    def test_equal_time_series(self):
        # use the same time series and see if the assert correctly says they are the same
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)

        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3,
                                     abs_tolerance=10.0)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=10.0)

    def test_unequal_time_series_rel_only(self):
        # slightly modify the "to be checked" time series and check that the assert is working
        # Try both when the mod is slightly less than the tolerance and slightly more
        # Only use relative tolerance
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        rel_tolerance = 0.1

        x_check_5_orig = float(x_check[5])
        tolerance = x_check_5_orig * rel_tolerance

        x_check[5] = x_check_5_orig + tolerance * 0.9  # should not cause an error since rel error
        # will be less than tolerance
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=rel_tolerance)

        x_check[5] = x_check_5_orig + tolerance * 1.1  # should cause an error since rel error will
        # be less than tolerance

        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         rel_tolerance=rel_tolerance)

        start_of_expected_errmsg = f"The following timeseries data are out of tolerance due to " \
                                   f"absolute (None) or relative ({rel_tolerance}) tolerance " \
                                   f"violations"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with {start_of_expected_errmsg} but "
                        f"instead was {actual_errmsg}")
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 2)
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 0)

    def test_unequal_time_series_abs_only(self):
        # slightly modify the "to be checked" time series and check that the assert is working
        # Try both when the mod is slightly less than the tolerance and slightly more
        # Only use absolute tolerance
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_check_5_orig = float(x_check[5])

        abs_tolerance = 10.0

        # should not cause an error since rel error will be less than tolerance
        x_check[5] = x_check_5_orig + abs_tolerance * 0.9
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=abs_tolerance)

        # should cause an error since rel error will be less than tolerance
        x_check[5] = x_check_5_orig + abs_tolerance * 1.1
        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         abs_tolerance=abs_tolerance)
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   f"absolute ({abs_tolerance}) or relative (None) tolerance " \
                                   "violations"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with {start_of_expected_errmsg} but "
                        f"instead was {actual_errmsg}")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 2)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 0)

    def test_unequal_time_series_abs_and_rel(self):
        # slightly modify the "to be checked" time series and check that the assert is working
        # Try both when the mod is slightly less than the tolerance and slightly more
        # Use both and absolute and relative tolerance.
        # Do the mods both in the range where the absolute tolerance is used (small values) and
        #   also relative tolerance (large values)
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_check_5_orig = float(x_check[5])

        abs_tolerance = 10.0
        rel_tolerance = 0.1

        # for < 100, uses the abs, x_check[5] is ~ 50
        # should not cause an error since rel error will be less than tolerance
        x_check[5] = x_check_5_orig + abs_tolerance * 0.9
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                     abs_tolerance=abs_tolerance,
                                     rel_tolerance=rel_tolerance
                                     )

        # should cause an error since rel error will be less than tolerance
        x_check[5] = x_check_5_orig + abs_tolerance * 1.1
        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         abs_tolerance=abs_tolerance,
                                         rel_tolerance=rel_tolerance
                                         )
        start_of_expected_errmsg = f"The following timeseries data are out of tolerance due to " \
                                   f"absolute"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with '{start_of_expected_errmsg}' but "
                        f"instead was '{actual_errmsg}'")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 2)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 2)

        # for > 100, uses the rel, x_check[15] is ~ 150
        x_check[5] = x_check_5_orig
        x_check_15_orig = float(x_check[15])
        tolerance = x_check_15_orig * rel_tolerance
        # should not cause an error since rel error will be less than tolerance
        x_check[15] = x_check_15_orig + tolerance * 0.9
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                     abs_tolerance=abs_tolerance,
                                     rel_tolerance=rel_tolerance
                                     )

        # should cause an error since rel error will be greater than tolerance
        x_check[15] = x_check_15_orig + tolerance * 1.1
        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         abs_tolerance=abs_tolerance,
                                         rel_tolerance=rel_tolerance
                                         )
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   "absolute"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with '{start_of_expected_errmsg}' but "
                        f"instead was '{actual_errmsg}'")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 2)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 2)

        # Combine the two cases where one data paint fails because of abs error and one because
        #   of rel error
        # should cause an error since rel error will be less than tolerance
        x_check[5] = x_check_5_orig + abs_tolerance * 1.1
        tolerance = x_check_15_orig * rel_tolerance
        # should cause an error since rel error will be less than tolerance
        x_check[15] = x_check_15_orig + tolerance * 1.1
        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         abs_tolerance=abs_tolerance,
                                         rel_tolerance=rel_tolerance
                                         )
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   "absolute"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with '{start_of_expected_errmsg}' but "
                        f"instead was '{actual_errmsg}'")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 3)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 3)

    def test_no_overlapping_time(self):
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 510.0, 1500.0, 0.0, 1000.0)
        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=1)
        expected_msg = f"There is no overlapping time between the two time series"
        actual_errmsg = str(e.exception)
        self.assertEqual(actual_errmsg, expected_msg)

    def test_with_overlapping_times(self):
        # checked time series shifted to the right
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 250.0, 750.0, 500.0, 1500.0)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

        # checked time series shifted to the left
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, -250.0, 250.0, -500.0, 500.0)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

        # checked time series is subset of reference
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, 100.0, 400.0, 200.0, 800.0)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

        # checked time series is superset of reference
        t_ref, x_ref = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check = create_linear_time_series(100, -500.0, 1000.0, -1000.0, 2000.0)
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

    def test_multi_dimensional_equal(self):
        t_ref, x_ref_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_ref, x_ref_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_ref = np.stack((x_ref_1, x_ref_2), axis=1)

        t_check, x_check_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_check, x_check_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_check = np.stack((x_check_1, x_check_2), axis=1)

        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

    def test_multi_dimensional_unequal(self):
        t_ref, x_ref_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_ref, x_ref_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_ref = np.stack((x_ref_1, x_ref_2), axis=1)

        # change the timeseries data values to be checked a little
        t_check, x_check_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 990.0)
        t_check, x_check_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 990.0)
        x_check = np.stack((x_check_1, x_check_2), axis=1)

        rel_tolerance = 1.0E-3

        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         rel_tolerance=rel_tolerance)
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   f"absolute (None) or relative ({rel_tolerance}) tolerance " \
                                   "violations"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with {start_of_expected_errmsg} but "
                        f"instead was {actual_errmsg}")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 0)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 199)

    def test_multi_dimensional_unequal_abs_and_rel(self):
        t_ref, x_ref_1 = create_linear_time_series(10, 0.0, 500.0, 0.0, 1000.0)
        t_ref, x_ref_2 = create_linear_time_series(10, 0.0, 500.0, 0.0, 1000.0)
        x_ref = np.stack((x_ref_1, x_ref_2), axis=1)

        # change the timeseries data values to be checked a little
        t_check, x_check_1 = create_linear_time_series(10, 0.0, 500.0, 0.0, 990.0)
        t_check, x_check_2 = create_linear_time_series(10, 0.0, 500.0, 0.0, 990.0)
        x_check = np.stack((x_check_1, x_check_2), axis=1)

        abs_tolerance = 1.0E-3
        rel_tolerance = 0.5E-5

        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         rel_tolerance=rel_tolerance, abs_tolerance=abs_tolerance)
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   f"absolute ({abs_tolerance}) or relative ({rel_tolerance}) " \
                                   "tolerance violations"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with {start_of_expected_errmsg} but "
                        f"instead was {actual_errmsg}")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 19)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 19)

    def test_multi_dimensional_with_overlapping_times(self):
        t_ref, x_ref_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_ref, x_ref_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_ref = np.stack((x_ref_1, x_ref_2), axis=1)

        t_check, x_check_1 = create_linear_time_series(100, 250.0, 750.0, 500.0, 1500.0)
        t_check, x_check_2 = create_linear_time_series(100, 250.0, 750.0, 500.0, 1500.0)
        x_check = np.stack((x_check_1, x_check_2), axis=1)

        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, rel_tolerance=1.0E-3)

    def test_multi_dimensional_unequal_with_overlapping_times(self):
        t_ref, x_ref_1 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        t_ref, x_ref_2 = create_linear_time_series(100, 0.0, 500.0, 0.0, 1000.0)
        x_ref = np.stack((x_ref_1, x_ref_2), axis=1)

        t_check, x_check_1 = create_linear_time_series(100, 250.0, 750.0, 500.0, 1000.0)
        t_check, x_check_2 = create_linear_time_series(100, 250.0, 750.0, 500.0, 1000.0)
        x_check = np.stack((x_check_1, x_check_2), axis=1)

        rel_tolerance = 1.0E-3

        with self.assertRaises(AssertionError) as e:
            assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check,
                                         rel_tolerance=rel_tolerance)
        start_of_expected_errmsg = "The following timeseries data are out of tolerance due to " \
                                   f"absolute (None) or relative ({rel_tolerance}) tolerance " \
                                   "violations"
        actual_errmsg = str(e.exception)
        self.assertTrue(actual_errmsg.startswith(start_of_expected_errmsg),
                        f"Error message expected to start with {start_of_expected_errmsg} but "
                        f"instead was {actual_errmsg}")
        self.assertEquals(actual_errmsg.count('>ABS_TOL'), 0)
        self.assertEquals(actual_errmsg.count('>REL_TOL'), 99)


@use_tempdirs
class TestAssertCasesEqual(unittest.TestCase):

    def tearDown(self):
        for file in ('p1.db', 'p2.db'):
            try:
                os.remove(file)
            except:
                pass

    def test_different_variables(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=0.0)
        ivc.add_output('b', val=1.0)
        ivc.add_output('c', val=2.0)
        ivc.add_output('x', val=3.0)
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('x', val=3.0)
        ivc.add_output('y', val=4.0)
        ivc.add_output('z', val=5.0)
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        expected = "\nrequire_same_vars=True but cases contain different variables.\nVariables in "\
                   "" \
                   "case1 but not in case2: ['a', 'b', 'c']\nVariables in case2 but not in " \
                   "case1: ['y', 'z']"

        self.assertEqual(str(e.exception), expected)

    def test_allow_different_variables(self):
        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=0.0)
        ivc.add_output('b', val=1.0)
        ivc.add_output('c', val=2.0)
        ivc.add_output('x', val=3.0)
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('x', val=3.0)
        ivc.add_output('y', val=4.0)
        ivc.add_output('z', val=5.0)
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        assert_cases_equal(c1, c2, require_same_vars=False)

    def test_different_shapes(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=np.array([0, 1, 2]))
        ivc.add_output('b', val=np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=np.eye(3, dtype=float))
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=3.0)
        ivc.add_output('b', val=np.array([0, 1, 2]))
        ivc.add_output('c', val=np.ones(3, dtype=float))
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        expected = "\nThe following variables have different shapes/sizes:\na has shape (3,) in " \
                   "case1 but shape (1,) in case2\nb has shape (2, 3) in case1 but shape (3,) in " \
                   "case2\nc has shape (3, 3) in case1 but shape (3,) in case2"

        self.assertEqual(str(e.exception), expected)

    def test_different_values(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=np.array([0, 1, 2]))
        ivc.add_output('b', val=np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=np.eye(3, dtype=float))
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=3 * np.array([0, 1, 2]))
        ivc.add_output('b', val=2 * np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=5 * np.eye(3, dtype=float))
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        expected = "\nThe following variables contain different values:\n" \
                   "var        max error       mean error\n" \
                   "--- ---------------- ----------------\n" \
                   "  a  4.000000000e+00  2.000000000e+00\n" \
                   "  b  5.000000000e+00  2.500000000e+00\n" \
                   "  c  4.000000000e+00  1.333333333e+00\n"

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        self.assertEqual(str(e.exception), expected)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import os
import unittest

import numpy as np
from scipy.interpolate import interp1d

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm


class BrachistochroneODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=9.80665 * np.ones(nn), desc='grav. acceleration', units='m/s/s')

        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')

        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s')

        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s')

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')

        self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                        units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='xdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='ydot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='check', wrt='theta', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta
        outputs['check'] = v / sin_theta

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        jacobian['vdot', 'g'] = cos_theta
        jacobian['vdot', 'theta'] = -g * sin_theta

        jacobian['xdot', 'v'] = sin_theta
        jacobian['xdot', 'theta'] = v * cos_theta

        jacobian['ydot', 'v'] = -cos_theta
        jacobian['ydot', 'theta'] = v * sin_theta

        jacobian['check', 'v'] = 1 / sin_theta
        jacobian['check', 'theta'] = -v * cos_theta / sin_theta**2


@use_tempdirs
class TestBrachistochroneIntegratedControl(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['phase0_sim.db']:
            if os.path.exists(filename):
                os.remove(filename)

    def test_brachistochrone_integrated_control_gauss_lobatto(self):
        import numpy as np
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='xdot', units='m')
        phase.add_state('y', fix_initial=True, fix_final=True, rate_source='ydot', units='m')
        phase.add_state('v', fix_initial=True, rate_source='vdot', units='m/s')
        phase.add_state('theta', targets='theta', fix_initial=False, rate_source='theta_dot')

        phase.add_control('theta_dot', units='deg/s', rate_continuity=True, shape=(1, ), lower=0, upper=60)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665, targets=['g'])

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p.set_val('phase0.states:x', phase.interp('x', ys=[0, 10]))
        p.set_val('phase0.states:y', phase.interp('y', ys=[10, 5]))
        p.set_val('phase0.states:v', phase.interp('v', ys=[0, 9.9]))
        p.set_val('phase0.states:theta', phase.interp('theta', ys=[5, 100.5]), units='deg')
        p.set_val('phase0.controls:theta_dot', phase.interp('theta_dot', ys=[50, 50]))

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        sim_out = phase.simulate(times_per_seg=20, rtol=1.0E-5)

        x_sol = p.get_val('phase0.timeseries.states:x')
        y_sol = p.get_val('phase0.timeseries.states:y')
        v_sol = p.get_val('phase0.timeseries.states:v')
        theta_sol = p.get_val('phase0.timeseries.states:theta')
        theta_dot_sol = p.get_val('phase0.timeseries.controls:theta_dot')
        time_sol = p.get_val('phase0.timeseries.time')

        x_sim = sim_out.get_val('phase0.timeseries.states:x')
        y_sim = sim_out.get_val('phase0.timeseries.states:y')
        v_sim = sim_out.get_val('phase0.timeseries.states:v')
        theta_sim = sim_out.get_val('phase0.timeseries.states:theta')
        theta_dot_sim = sim_out.get_val('phase0.timeseries.controls:theta_dot')
        time_sim = sim_out.get_val('phase0.timeseries.time')

        x_interp = interp1d(time_sim[:, 0], x_sim[:, 0])
        y_interp = interp1d(time_sim[:, 0], y_sim[:, 0])
        v_interp = interp1d(time_sim[:, 0], v_sim[:, 0])
        theta_interp = interp1d(time_sim[:, 0], theta_sim[:, 0])
        theta_dot_interp = interp1d(time_sim[:, 0], theta_dot_sim[:, 0])

        assert_near_equal(x_interp(time_sol), x_sol, tolerance=1.0E-4)
        assert_near_equal(y_interp(time_sol), y_sol, tolerance=1.0E-4)
        assert_near_equal(v_interp(time_sol), v_sol, tolerance=1.0E-4)
        assert_near_equal(theta_interp(time_sol), theta_sol, tolerance=1.0E-4)
        assert_near_equal(theta_dot_interp(time_sol), theta_dot_sol, tolerance=1.0E-4)

    def test_brachistochrone_integrated_control_radau_ps(self):
        import numpy as np
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=dm.Radau(num_segments=10))

        traj.add_phase('phase0', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='xdot', units='m')
        phase.add_state('y', fix_initial=True, fix_final=True, rate_source='ydot', units='m')
        phase.add_state('v', fix_initial=True, rate_source='vdot', units='m/s')
        phase.add_state('theta', targets='theta', fix_initial=False, rate_source='theta_dot')

        phase.add_control('theta_dot', units='deg/s', rate_continuity=True, shape=(1, ), lower=0, upper=60)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665, targets=['g'])

        phase.set_simulate_options(rtol=1.0E-8, atol=1.0E-8)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 2.0

        p.set_val('traj.phase0.states:x', phase.interp('x', ys=[0, 10]))
        p.set_val('traj.phase0.states:y', phase.interp('y', ys=[10, 5]))
        p.set_val('traj.phase0.states:v', phase.interp('v', ys=[0, 9.9]))
        p.set_val('traj.phase0.states:theta', phase.interp('theta', ys=[5, 100.5]), units='deg')
        p.set_val('traj.phase0.controls:theta_dot', phase.interp('theta_dot', ys=[50, 50]))

        # Solve for the optimal trajectory
        dm.run_problem(p, simulate=True, make_plots=True)

        sol_case = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case = om.CaseReader('dymos_simulation.db').get_case('final')

        # Test the results
        assert_near_equal(sol_case.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        x_sol = sol_case.get_val('traj.phase0.timeseries.states:x')
        y_sol = sol_case.get_val('traj.phase0.timeseries.states:y')
        v_sol = sol_case.get_val('traj.phase0.timeseries.states:v')
        theta_sol = sol_case.get_val('traj.phase0.timeseries.states:theta')
        theta_dot_sol = sol_case.get_val('traj.phase0.timeseries.controls:theta_dot')
        time_sol = sol_case.get_val('traj.phase0.timeseries.time')

        x_sim = sim_case.get_val('traj.phase0.timeseries.states:x')
        y_sim = sim_case.get_val('traj.phase0.timeseries.states:y')
        v_sim = sim_case.get_val('traj.phase0.timeseries.states:v')
        theta_sim = sim_case.get_val('traj.phase0.timeseries.states:theta')
        theta_dot_sim = sim_case.get_val('traj.phase0.timeseries.controls:theta_dot')
        time_sim = sim_case.get_val('traj.phase0.timeseries.time')

        x_interp = interp1d(time_sim[:, 0], x_sim[:, 0])
        y_interp = interp1d(time_sim[:, 0], y_sim[:, 0])
        v_interp = interp1d(time_sim[:, 0], v_sim[:, 0])
        theta_interp = interp1d(time_sim[:, 0], theta_sim[:, 0])
        theta_dot_interp = interp1d(time_sim[:, 0], theta_dot_sim[:, 0])

        assert_near_equal(x_interp(time_sol), x_sol, tolerance=1.0E-4)
        assert_near_equal(y_interp(time_sol), y_sol, tolerance=1.0E-4)
        assert_near_equal(v_interp(time_sol), v_sol, tolerance=1.0E-4)
        assert_near_equal(theta_interp(time_sol), theta_sol, tolerance=1.0E-4)
        assert_near_equal(theta_dot_interp(time_sol), theta_dot_sol, tolerance=1.0E-4)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


def lgr(n, include_endpoint=False, tol=1.0E-15):
    """
    Returns the Legendre-Gauss-Radau nodes and weights for a Jacobi Polynomial with n abscissae.

    Parameters
    ----------
    n : int
        The number of LGR nodes are to be returned.  If include_endpoint=True we return n+1 values.
    include_endpoint : bool
        If True return the non-free abscissa at the right endpoint of the interval [-1, 1].
        The weight associated with this endpoint is 0.
    tol : float
        The tolerance to which the location of the nodes should be converged.

    Returns
    -------
    numpy.array
        An array of the LGR nodes for a polynomial of the given order.
    numpy.array
        An array of the corresponding LGR weights at the nodes in x.

    Notes
    -----
    The LGR nodes are the roots of

        .. math::
            (P_N(x)+P_{N+1}(x))/(x+1).

    References
    ----------
    C. Canuto, M. Y. Hussaini, A. Quarteroni, T. A. Tang, "Spectral Methods
    in Fluid Dynamics," Section 2.3. Springer-Verlag 1987

    F. B. Hildebrand , "Introduction to Numerical Analysis," Section 8.11
    Dover 1987

    The nodes are on the range [-1, 1).

    Based on the routine written by Greg von Winckel (License follows)

    Copyright (c) 2004, Greg von Winckel
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      * Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
      * Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in
        the documentation and/or other materials provided with the distribution

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR aux_outputs
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    n1 = n
    n = n - 1

    # The Chebyshev-Gauss-Radau nodes serve as the initial guess.
    x = -np.cos(2.0 * np.pi * np.arange(0, n1) / (2 * n + 1))

    # The Legendre vandermonde Matrix
    P = np.zeros((n1, n1 + 1))

    # Compute P_(N) using the recursion relation
    # Copute its first and second derivatives and
    # update x using the Newton-Raphson method.
    xold = np.ones_like(x)

    # Free abscissae
    free = np.arange(1, n1)

    for i in range(100):
        if np.all(np.abs(x - xold) <= tol):
            break

        xold[:] = x

        P[0, :] = np.power(-1.0, np.arange(0, n1 + 1))
        P[free, 0] = 1.0
        P[free, 1] = x[free]

        for k in range(1, n1):
            P[free, k + 1] = (
                (2 * (k + 1) - 1) * x[free] * P[free, k] -
                ((k + 1) - 1) * P[free, k - 1]
            ) / (k + 1)

        f = ((1.0 - xold[free]) / n1) * (P[free, n] + P[free, n1])
        fprime = (P[free, n] - P[free, n1])

        x[free] = xold[free] - f / fprime

    else:
        raise RuntimeError(f'Failed to converge LGR nodes for order {n}')

    # Compute the weights
    w = np.zeros(n1)
    w[0] = 2.0 / n1**2
    w[free] = (1.0 - x[free]) / (n1 * P[free, n])**2

    # Tack on the endpoint if requested
    if include_endpoint:
        x = np.concatenate([x, [1.0]])
        w = np.concatenate([w, [0.0]])

    return x, w



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from dymos.utils.testing_utils import assert_check_partials

import dymos as dm
from dymos.models.eom import FlightPathEOM2D

OPTIMIZER = 'SLSQP'
SHOW_PLOTS = False


class _CannonballODE(FlightPathEOM2D):
    pass


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestFlightPathEOM2D(unittest.TestCase):

    def setUp(self):
        self.p = om.Problem(model=om.Group())

        self.p.driver = om.pyOptSparseDriver()
        self.p.driver.options['optimizer'] = OPTIMIZER
        if OPTIMIZER == 'SNOPT':
            self.p.driver.opt_settings['Major iterations limit'] = 50
            self.p.driver.opt_settings['iSumm'] = 6
            self.p.driver.opt_settings['Verify level'] = 3

        phase = dm.Phase(ode_class=_CannonballODE,
                         transcription=dm.GaussLobatto(num_segments=5, order=3, compressed=False))

        self.p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(10, 20), units='s')

        phase.add_state('r', rate_source='r_dot', units='m',
                        fix_initial=True, fix_final=False,
                        scaler=0.001, defect_scaler=0.001)

        phase.add_state('h', rate_source='h_dot', units='m',
                        fix_initial=True, fix_final=True,
                        scaler=0.001, defect_scaler=0.001)

        phase.add_state('v', rate_source='v_dot', targets='v', units='m/s',
                        fix_initial=True, fix_final=False,
                        scaler=0.01, defect_scaler=0.01)

        phase.add_state('gam', rate_source='gam_dot', targets='gam', units='rad',
                        fix_final=False, scaler=1.0, defect_scaler=1.0)

        # Maximize final range by varying initial flight path angle
        phase.add_objective('r', loc='final', scaler=-0.01)

    def test_cannonball_simulate(self):
        self.p.setup()

        v0 = 100.0
        gam0 = np.radians(45)
        g = 9.80665
        t_duration = 2 * v0 * np.sin(gam0) / g

        phase = self.p.model.phase0

        self.p['phase0.t_initial'] = 0.0
        self.p['phase0.t_duration'] = t_duration

        self.p['phase0.states:r'] = phase.interp('r', [0, 700.0])
        self.p['phase0.states:h'] = phase.interp('h', [0, 0])
        self.p['phase0.states:v'] = phase.interp('v', [v0, v0])
        self.p['phase0.states:gam'] = phase.interp('gam', [gam0, -gam0])

        self.p.run_model()

        exp_out = phase.simulate()

        assert_near_equal(exp_out.get_val('phase0.timeseries.states:h', units='km')[-1], 0.0,
                          tolerance=0.001)
        assert_near_equal(exp_out.get_val('phase0.timeseries.states:r')[-1], v0**2 / g,
                          tolerance=0.001)
        assert_near_equal(exp_out.get_val('phase0.timeseries.states:gam')[-1], -gam0,
                          tolerance=0.001)
        assert_near_equal(exp_out.get_val('phase0.timeseries.states:v')[-1], v0,
                          tolerance=0.001)

    def test_cannonball_max_range(self):
        self.p.setup(force_alloc_complex=True)

        v0 = 100.0
        gam0 = np.radians(45.0)
        g = 9.80665
        t_duration = 10.0

        phase = self.p.model.phase0

        self.p['phase0.t_initial'] = 0.0
        self.p['phase0.t_duration'] = t_duration

        self.p['phase0.states:r'] = phase.interp('r', [0, v0 * np.cos(gam0) * t_duration])
        self.p['phase0.states:h'] = phase.interp('h', [0, 0])
        self.p['phase0.states:v'] = phase.interp('v', [v0, v0])
        self.p['phase0.states:gam'] = phase.interp('gam', [gam0, -gam0])

        self.p.run_driver()

        with np.printoptions(linewidth=1024):
            dm.options['include_check_partials'] = True
            self.p.check_partials(method='cs')

        exp_out = phase.simulate(times_per_seg=None)

        assert_near_equal(exp_out.get_val('phase0.timeseries.states:r')[-1], v0**2 / g,
                          tolerance=0.001)
        assert_near_equal(exp_out.get_val('phase0.timeseries.states:gam')[-1], -np.radians(45),
                          tolerance=0.001)
        assert_near_equal(exp_out.get_val('phase0.timeseries.states:v')[-1], v0,
                          tolerance=0.001)

    def test_partials(self):
        self.p.setup(force_alloc_complex=True)

        v0 = 100.0
        gam0 = np.radians(30.0)
        t_duration = 10.0

        phase = self.p.model.phase0

        self.p['phase0.t_initial'] = 0.0
        self.p['phase0.t_duration'] = t_duration

        self.p['phase0.states:r'] = phase.interp('r', [0, v0 * np.cos(gam0) * t_duration])
        self.p['phase0.states:h'] = phase.interp('h', [0, 0])
        self.p['phase0.states:v'] = phase.interp('v', [v0, v0])
        self.p['phase0.states:gam'] = phase.interp('gam', [gam0, -gam0])

        self.p.run_model()

        cpd = self.p.check_partials(compact_print=True, method='cs')

        assert_check_partials(cpd)

if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from numbers import Number
from collections.abc import Iterable

import openmdao.api as om
from ..utils.misc import _unspecified


class LinkageOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for Phase linkages.
    """

    def __init__(self):
        super().__init__()

        self.declare(name='phase_a', types=str,
                     desc='name of the first phase in the linkage')

        self.declare(name='phase_b', types=str,
                     desc='name of the second phase in the linkage')

        self.declare(name='var_a', types=str,
                     desc='name of the first variable in the linkage')

        self.declare(name='var_b', types=str,
                     desc='name of the second variable in the linkage')

        self.declare(name='loc_a', values=('initial', 'final'),
                     desc='location of the first variable in the linkage (\'initial\' or \'final\')')

        self.declare(name='loc_b', values=('initial', 'final'),
                     desc='location of the second variable in the linkage (\'initial\' or \'final\')')

        self.declare(name='sign_a', types=Number, default=1.0,
                     desc='sign of the first variable in the linkage (\'initial\' or \'final\')')

        self.declare(name='sign_b', types=Number, default=-1.0,
                     desc='sign of the second variable in the linkage (\'initial\' or \'final\')')

        self.declare(name='units_a', default=_unspecified,
                     allow_none=True, desc='units in which the first variable is defined')

        self.declare(name='units_b', default=_unspecified,
                     allow_none=True, desc='units in which the second variable is defined')

        self.declare(name='units', default=_unspecified,
                     allow_none=True, desc='units in which the linkage constraint is defined')

        self.declare(name='shape', types=Iterable, allow_none=True, default=None,
                     desc='shape of the state variable, as determined by introspection')

        self.declare(name='lower', types=(Iterable, Number), default=None,
                     allow_none=True, desc='Lower bound of the resulting constraint.')

        self.declare(name='upper',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Upper bound of the resulting constraint.')

        self.declare(name='equals',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Equality constraint value of the resulting constraint.')

        self.declare(name='scaler',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Scaler of the resulting constraint.')

        self.declare(name='adder',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Adder of the resulting constraint.')

        self.declare(name='ref0',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Zero-reference of the resulting constraint.')

        self.declare(name='ref',
                     types=(Iterable, Number), default=None,
                     allow_none=True, desc='Unit-reference of the resulting constraint.')

        self.declare(name='linear', types=bool, default=False,
                     desc='If True, treat the resulting constraint as a linear constraint. This '
                          'option should only be applied to linked design variables and time.')

        self.declare(name='connected', types=bool, default=False,
                     desc='If True, this linkage is handled as a direct connection rather than'
                          'by a constraint.  This option only applies to time and states.')

        self.declare(name='constraint_name', types=str, default=None, allow_none=True,
                     desc='optional alternative constraint name to override name conflicts.')

        self._src_a = None
        self._src_b = None
        self._input_a = None
        self._input_b = None
        self._output = None



import unittest


class TestIntrospection(unittest.TestCase):

    def test_filter_outputs(self):
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        import openmdao.api as om

        from dymos.utils.introspection import filter_outputs

        p = om.Problem()
        p.model.add_subsystem('ode', MinTimeClimbODE(num_nodes=1))

        p.setup()

        outputs = filter_outputs(['atmos.*', 'aero.*'], p.model.ode)

        expected = 'atmos.drhos_dh atmos.temp atmos.pres atmos.rho atmos.viscosity atmos.sos ' \
                   'aero.mach aero.CD0 aero.kappa aero.CLa aero.CL aero.CD aero.q aero.f_lift ' \
                   'aero.f_drag'

        self.assertSetEqual(set(outputs.keys()), set(expected.split()))

    def test_filter_outputs_with_dict(self):
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        import openmdao.api as om

        from dymos.utils.introspection import get_promoted_vars, filter_outputs

        p = om.Problem()
        p.model.add_subsystem('ode', MinTimeClimbODE(num_nodes=1))

        p.setup()

        outputs = filter_outputs(['atmos.*', 'aero.*'], get_promoted_vars(p.model.ode, 'output'))

        expected = 'atmos.drhos_dh atmos.temp atmos.pres atmos.rho atmos.viscosity atmos.sos ' \
                   'aero.mach aero.CD0 aero.kappa aero.CLa aero.CL aero.CD aero.q aero.f_lift ' \
                   'aero.f_drag'

        self.assertSetEqual(set(outputs.keys()), set(expected.split()))



import unittest
import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials

from dymos.models.atmosphere.atmos_1976 import USatm1976Comp, USatm1976Data, _build_akima_coefs


class TestAtmosphere(unittest.TestCase):

    def test_atmos_comp_geopotential(self):
        n = USatm1976Data.alt.size

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output(name='alt', val=USatm1976Data.alt, units='ft')

        p.model.add_subsystem('atmos', subsys=USatm1976Comp(num_nodes=n))
        p.model.connect('alt', 'atmos.h')

        p.setup(force_alloc_complex=True)
        p.run_model()

        T = p.get_val('atmos.temp', units='degR')
        P = p.get_val('atmos.pres', units='psi')
        rho = p.get_val('atmos.rho', units='slug/ft**3')
        sos = p.get_val('atmos.sos', units='ft/s')

        assert_near_equal(T, USatm1976Data.T, tolerance=1.0E-4)
        assert_near_equal(P, USatm1976Data.P, tolerance=1.0E-4)
        assert_near_equal(rho, USatm1976Data.rho, tolerance=1.0E-4)
        assert_near_equal(sos, USatm1976Data.a, tolerance=1.0E-4)

        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)

    def test_atmos_comp_geodetic(self):
        n = USatm1976Data.alt.size

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output(name='alt', val=USatm1976Data.alt, units='ft')

        p.model.add_subsystem('atmos', subsys=USatm1976Comp(num_nodes=n, h_def='geodetic'))
        p.model.connect('alt', 'atmos.h')

        p.setup(force_alloc_complex=True)

        h = USatm1976Data.alt * 0.3048  # altitude data in meters
        R0 = 6_356_766  # US 1976 std atm R0 in m
        p.set_val('alt', R0 / (R0 - h) * h, units='m')  # US 1976 std atm geopotential altitude to geodetic (m)

        p.run_model()

        T = p.get_val('atmos.temp', units='degR')
        P = p.get_val('atmos.pres', units='psi')
        rho = p.get_val('atmos.rho', units='slug/ft**3')
        sos = p.get_val('atmos.sos', units='ft/s')

        assert_near_equal(T, USatm1976Data.T, tolerance=1.0E-4)
        assert_near_equal(P, USatm1976Data.P, tolerance=1.0E-4)
        assert_near_equal(rho, USatm1976Data.rho, tolerance=1.0E-4)
        assert_near_equal(sos, USatm1976Data.a, tolerance=1.0E-4)

        with np.printoptions(linewidth=100000):
            cpd = p.check_partials(method='cs')
        assert_check_partials(cpd)

    def test_atmos_build_akima_coefs(self):
        coeff_data = _build_akima_coefs(out_stream=None)

        for key, vals in coeff_data.items():
            saved_vals = getattr(USatm1976Data, key.split('.')[-1])
            assert_near_equal(vals, saved_vals, tolerance=1.0E-15)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs
from dymos.examples.robertson_problem.doc.robertson_ode import RobertsonODE


@use_tempdirs
class TestRobertsonProblemForDocs(unittest.TestCase):

    def robertson_problem(self, t_final=1.0):

        import openmdao.api as om
        import dymos as dm

        #
        # Initialize the Problem
        #
        p = om.Problem(model=om.Group())

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=RobertsonODE,
                                        transcription=dm.GaussLobatto(num_segments=50)
                                        ))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, fix_duration=True)

        phase.add_state('x0', fix_initial=True, fix_final=False, rate_source='xdot', targets='x')
        phase.add_state('y0', fix_initial=True, fix_final=False, rate_source='ydot', targets='y')
        phase.add_state('z0', fix_initial=True, fix_final=False, rate_source='zdot', targets='z')

        #
        # Setup the Problem
        #
        p.setup(check=True)

        #
        # Set the initial values
        #
        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = t_final

        p.set_val('traj.phase0.states:x0', phase.interp('x0', ys=[1.0, 0.7]))
        p.set_val('traj.phase0.states:y0', phase.interp('y0', ys=[0.0, 1e-5]))
        p.set_val('traj.phase0.states:z0', phase.interp('z0', ys=[0.0, 0.3]))

        return p

    def test_robertson_problem_for_docs(self):

        import openmdao.api as om
        from dymos.utils.testing_utils import assert_check_partials
        from openmdao.utils.assert_utils import assert_near_equal
        import matplotlib.pyplot as plt

        num_nodes = 3

        p = om.Problem(model=om.Group())

        p.model.add_subsystem('ode', RobertsonODE(num_nodes=num_nodes), promotes=['*'])

        p.setup(force_alloc_complex=True)

        p.set_val('x', [10., 100., 1000.])
        p.set_val('y', [1, 0.1, 0.01])
        p.set_val('z', [1., 2., 3.])

        p.run_model()
        cpd = p.check_partials(method='cs', compact_print=True)

        assert_check_partials(cpd)

        assert_near_equal(p.get_val('xdot'), [9999.6, 1996., 260.])
        assert_near_equal(p.get_val('ydot'), [-3.00099996E7, -3.01996E5, -3.26E3])
        assert_near_equal(p.get_val('zdot'), [3.0E7, 3.0E5, 3.0E3])

        # just set up the problem, test it elsewhere
        p = self.robertson_problem(t_final=40)

        p.run_model()

        p_sim = p.model.traj.simulate(method='LSODA')

        assert_near_equal(p_sim.get_val('traj.phase0.timeseries.states:x0')[-1], 0.71583161, tolerance=1E-5)
        assert_near_equal(p_sim.get_val('traj.phase0.timeseries.states:y0')[-1], 9.18571144e-06, tolerance=1E-1)
        assert_near_equal(p_sim.get_val('traj.phase0.timeseries.states:z0')[-1], 0.2841592, tolerance=1E-5)

        t_sim = p_sim.get_val('traj.phase0.timeseries.time')

        states = ['x0', 'y0', 'z0']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            axes[i].plot(t_sim, p_sim.get_val(f'traj.phase0.timeseries.states:{state}'), '-')
            axes[i].set_ylabel(state[0])
        axes[-1].set_xlabel('time (s)')
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    unittest.main()



from collections import defaultdict
from fnmatch import filter
import warnings

import numpy as np
import openmdao.api as om
from openmdao.utils.om_warnings import issue_warning

from .pseudospectral_base import PseudospectralBase
from ..common import RadauPSContinuityComp
from ...utils.misc import get_rate_units, _unspecified
from ...utils.introspection import get_promoted_vars, get_targets, get_source_metadata
from ...utils.indexing import get_src_indices_by_row
from ..grid_data import GridData


class Radau(PseudospectralBase):
    """
    Radau Pseudospectral Method Transcription.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    References
    ----------
    Garg, Divya et al. "Direct Trajectory Optimization and Costate Estimation of General Optimal
    Control Problems Using a Radau Pseudospectral Method." American Institute of Aeronautics
    and Astronautics, 2009.
    """
    def __init__(self, **kwargs):
        super(Radau, self).__init__(**kwargs)
        self._rhs_source = 'rhs_all'

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        self.grid_data = GridData(num_segments=self.options['num_segments'],
                                  transcription='radau-ps',
                                  transcription_order=self.options['order'],
                                  segment_ends=self.options['segment_ends'],
                                  compressed=self.options['compressed'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        This method assumes that target introspection has already been performed by the phase and thus
        options['targets'], options['time_phase_targets'], options['t_initial_targets'],
        and options['t_duration_targets'] are all correctly populated.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).configure_time(phase)
        options = phase.time_options
        ode = phase._get_subsystem(self._rhs_source)
        ode_inputs = get_promoted_vars(ode, iotypes='input')

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, targets, dynamic in [('time', options['targets'], True),
                                       ('time_phase', options['time_phase_targets'], True)]:
            if targets:
                src_idxs = self.grid_data.subset_node_indices['all'] if dynamic else None
                phase.connect(name, [f'rhs_all.{t}' for t in targets], src_indices=src_idxs,
                              flat_src_indices=True if dynamic else None)

        for name, targets in [('t_initial', options['t_initial_targets']),
                              ('t_duration', options['t_duration_targets'])]:
            for t in targets:
                shape = ode_inputs[t]['shape']

                if shape == (1,):
                    src_idxs = None
                    flat_src_idxs = None
                    src_shape = None
                else:
                    src_idxs = np.zeros(self.grid_data.subset_num_nodes['all'])
                    flat_src_idxs = True
                    src_shape = (1,)

                phase.promotes('rhs_all', inputs=[(t, name)], src_indices=src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)
            if targets:
                phase.set_input_defaults(name=name,
                                         val=np.ones((1,)),
                                         units=options['units'])

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).configure_controls(phase)

        for name, options in phase.control_options.items():
            if options['targets']:
                phase.connect(f'control_values:{name}', [f'rhs_all.{t}' for t in options['targets']])

            if options['rate_targets']:
                phase.connect(f'control_rates:{name}_rate',
                              [f'rhs_all.{t}' for t in options['rate_targets']])

            if options['rate2_targets']:
                phase.connect(f'control_rates:{name}_rate2',
                              [f'rhs_all.{t}' for t in options['rate2_targets']])

    def configure_polynomial_controls(self, phase):
        """
        Configure the inputs/outputs for the polynomial controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).configure_polynomial_controls(phase)

        ode_inputs = get_promoted_vars(self._get_ode(phase), 'input')

        for name, options in phase.polynomial_control_options.items():
            targets = get_targets(ode=ode_inputs, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'polynomial_control_values:{name}',
                              [f'rhs_all.{t}' for t in targets])

            targets = get_targets(ode=phase.rhs_all, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              [f'rhs_all.{t}' for t in targets])

            targets = get_targets(ode=phase.rhs_all, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'rhs_all.{t}' for t in targets])

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).setup_ode(phase)

        ODEClass = phase.options['ode_class']
        grid_data = self.grid_data

        kwargs = phase.options['ode_init_kwargs']
        phase.add_subsystem('rhs_all',
                            subsys=ODEClass(num_nodes=grid_data.subset_num_nodes['all'],
                                            **kwargs))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).configure_ode(phase)

        grid_data = self.grid_data
        map_input_indices_to_disc = grid_data.input_maps['state_input_to_disc']
        ode_inputs = get_promoted_vars(phase.rhs_all, 'input')

        for name, options in phase.state_options.items():

            targets = get_targets(ode_inputs, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'states:{name}',
                              [f'rhs_all.{tgt}' for tgt in targets],
                              src_indices=om.slicer[map_input_indices_to_disc, ...])

    def setup_defects(self, phase):
        """
        Create the continuity_comp to house the defects.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).setup_defects(phase)

        if any(self._requires_continuity_constraints(phase)):
            phase.add_subsystem('continuity_comp',
                                RadauPSContinuityComp(grid_data=self.grid_data,
                                                      state_options=phase.state_options,
                                                      control_options=phase.control_options,
                                                      time_units=phase.time_options['units']))

    def configure_defects(self, phase):
        """
        Configure the continuity_comp and connect the collocation constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Radau, self).configure_defects(phase)

        for name in phase.state_options:
            phase.connect(f'state_interp.staterate_col:{name}',
                          f'collocation_constraint.f_approx:{name}')

            rate_src_path, src_idxs = self._get_rate_source_path(name, 'col', phase)
            phase.connect(rate_src_path,
                          f'collocation_constraint.f_computed:{name}',
                          src_indices=src_idxs)

        any_state_cnty, any_control_cnty, any_control_rate_cnty = self._requires_continuity_constraints(phase)

        if any((any_state_cnty, any_control_cnty, any_control_rate_cnty)):
            phase._get_subsystem('continuity_comp').configure_io()

    def _get_rate_source_path(self, state_name, nodes, phase):
        """
        Return the rate source location and indices for a given state name.

        Parameters
        ----------
        state_name : str
            Name of the state.
        nodes : str
            One of ['col', 'all'].
        phase : dymos.Phase
            Phase object containing the rate source.

        Returns
        -------
        str
            Path to the rate source.
        ndarray
            Array of source indices.
        """
        gd = self.grid_data
        try:
            var = phase.state_options[state_name]['rate_source']
        except RuntimeError:
            raise ValueError(f"state '{state_name}' in phase '{ phase.name}' was not given a "
                             "rate_source")

        # Note the rate source must be shape-compatible with the state
        var_type = phase.classify_var(var)

        # Determine the path to the variable
        if var_type == 'time':
            rate_path = 'time'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'time_phase':
            rate_path = 'time_phase'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'state':
            rate_path = f'states:{var}'
            # Find the state_input indices which occur at segment endpoints, and repeat them twice
            state_input_idxs = gd.subset_node_indices['state_input']
            repeat_idxs = np.ones_like(state_input_idxs)
            if self.options['compressed']:
                segment_end_idxs = gd.subset_node_indices['segment_ends'][1:-1]
                # Repeat nodes that are on segment bounds (but not the first or last nodes in the phase)
                nodes_to_repeat = list(set(state_input_idxs).intersection(segment_end_idxs))
                # Now find these nodes in the state input indices
                idxs_of_ntr_in_state_inputs = np.where(np.in1d(state_input_idxs, nodes_to_repeat))[0]
                # All state input nodes are used once, but nodes_to_repeat are used twice
                repeat_idxs[idxs_of_ntr_in_state_inputs] = 2
            # Now we have a way of mapping the state input indices to all nodes
            map_input_node_idxs_to_all = np.repeat(np.arange(gd.subset_num_nodes['state_input'],
                                                             dtype=int), repeats=repeat_idxs)
            # Now select the subset of nodes we want to use.
            node_idxs = map_input_node_idxs_to_all[gd.subset_node_indices[nodes]]
        elif var_type == 'indep_control':
            rate_path = f'control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'input_control':
            rate_path = f'control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'control_rate':
            control_name = var[:-5]
            rate_path = f'control_rates:{control_name}_rate'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            rate_path = f'control_rates:{control_name}_rate2'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'indep_polynomial_control':
            rate_path = f'polynomial_control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'input_polynomial_control':
            rate_path = f'polynomial_control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'polynomial_control_rate':
            control_name = var[:-5]
            rate_path = f'polynomial_control_rates:{control_name}_rate'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'polynomial_control_rate2':
            control_name = var[:-6]
            rate_path = f'polynomial_control_rates:{control_name}_rate2'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'parameter':
            rate_path = f'parameter_vals:{var}'
            dynamic = not phase.parameter_options[var]['static_target']
            if dynamic:
                node_idxs = np.zeros(gd.subset_num_nodes[nodes], dtype=int)
            else:
                node_idxs = np.zeros(1, dtype=int)
        else:
            # Failed to find variable, assume it is in the ODE
            rate_path = f'rhs_all.{var}'
            node_idxs = gd.subset_node_indices[nodes]

        src_idxs = om.slicer[node_idxs, ...]

        return rate_path, src_idxs

    def _get_timeseries_var_source(self, var, output_name, phase):
        """
        Return the source path and indices for a given variable to be connected to a timeseries.

        Parameters
        ----------
        var : str
            Name of the timeseries variable whose source is desired.
        output_name : str
            Name of the timeseries output whose source is desired.
        phase : dymos.Phase
            Phase object containing the variable, either as state, time, control, etc., or as an ODE output.

        Returns
        -------
        meta : dict
            Metadata pertaining to the variable at the given path. This dict contains 'src' (the path to the
            timeseries source), 'src_idxs' (an array of the
            source indices), 'units' (the units of the source variable), and 'shape' (the shape of the variable at
            a given node).
        """
        gd = self.grid_data
        var_type = phase.classify_var(var)
        time_units = phase.time_options['units']

        transcription = phase.options['transcription']
        ode = transcription._get_ode(phase)
        ode_outputs = get_promoted_vars(ode, 'output')

        # The default for node_idxs, applies to everything except states and parameters.
        node_idxs = gd.subset_node_indices['all']

        meta = {}

        # Determine the path to the variable
        if var_type == 'time':
            path = 'time'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'time_phase':
            path = 'time_phase'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'state':
            path = f'states:{var}'
            src_units = phase.state_options[var]['units']
            src_shape = phase.state_options[var]['shape']

            # Find the state_input indices which occur at segment endpoints, and repeat them twice
            state_input_idxs = gd.subset_node_indices['state_input']
            repeat_idxs = np.ones_like(state_input_idxs)
            if self.options['compressed']:
                segment_end_idxs = gd.subset_node_indices['segment_ends'][1:-1]
                # Repeat nodes that are on segment bounds (but not the first or last nodes in the phase)
                nodes_to_repeat = list(set(state_input_idxs).intersection(set(segment_end_idxs)))
                # Now find these nodes in the state input indices
                idxs_of_ntr_in_state_inputs = np.where(np.in1d(state_input_idxs, nodes_to_repeat))[0]
                # All state input nodes are used once, but nodes_to_repeat are used twice
                repeat_idxs[idxs_of_ntr_in_state_inputs] = 2
            # Now we have a way of mapping the state input indices to all nodes
            map_input_node_idxs_to_all = np.repeat(np.arange(gd.subset_num_nodes['state_input'],
                                                             dtype=int), repeats=repeat_idxs)
            # Now select the subset of nodes we want to use.
            node_idxs = map_input_node_idxs_to_all[gd.subset_node_indices['all']]
        elif var_type in ['indep_control', 'input_control']:
            path = f'control_values:{var}'
            src_units = phase.control_options[var]['units']
            src_shape = phase.control_options[var]['shape']
        elif var_type == 'control_rate':
            control_name = var[:-5]
            path = f'control_rates:{control_name}_rate'
            control_name = var[:-5]
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=1)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            path = f'control_rates:{control_name}_rate2'
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=2)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type in ['indep_polynomial_control', 'input_polynomial_control']:
            path = f'polynomial_control_values:{var}'
            src_units = phase.polynomial_control_options[var]['units']
            src_shape = phase.polynomial_control_options[var]['shape']
        elif var_type == 'polynomial_control_rate':
            control_name = var[:-5]
            path = f'polynomial_control_rates:{control_name}_rate'
            control = phase.polynomial_control_options[control_name]
            src_units = get_rate_units(control['units'], time_units, deriv=1)
            src_shape = control['shape']
        elif var_type == 'polynomial_control_rate2':
            control_name = var[:-6]
            path = f'polynomial_control_rates:{control_name}_rate2'
            control = phase.polynomial_control_options[control_name]
            src_units = get_rate_units(control['units'], time_units, deriv=2)
            src_shape = control['shape']
        elif var_type == 'parameter':
            path = f'parameter_vals:{var}'
            # Timeseries are never a static_target
            node_idxs = np.zeros(gd.subset_num_nodes['all'], dtype=int)
            src_units = phase.parameter_options[var]['units']
            src_shape = phase.parameter_options[var]['shape']
        else:
            # Failed to find variable, assume it is in the ODE
            path = f'rhs_all.{var}'
            meta = get_source_metadata(ode_outputs, src=var)
            src_shape = meta['shape']
            src_units = meta['units']
            src_tags = meta['tags']
            if 'dymos.static_output' in src_tags:
                raise RuntimeError(f'ODE output {var} is tagged with "dymos.static_output" and cannot be a timeseries output.')

        src_idxs = om.slicer[node_idxs, ...]

        meta['src'] = path
        meta['src_idxs'] = src_idxs
        meta['units'] = src_units
        meta['shape'] = src_shape

        return meta

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            Parameter name.
        phase : dymos.Phase
            The phase object to which this transcription instance applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []

        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            if not options['static_target']:
                src_idxs_raw = np.zeros(self.grid_data.subset_num_nodes['all'], dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])
                if options['shape'] == (1,):
                    src_idxs = src_idxs.ravel()
            else:
                src_idxs_raw = np.zeros(1, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])
                src_idxs = np.squeeze(src_idxs, axis=0)

            rhs_all_tgts = [f'rhs_all.{t}' for t in options['targets']]
            connection_info.append((rhs_all_tgts, (src_idxs,)))

        return connection_info

    def _requires_continuity_constraints(self, phase):
        """
        Tests whether state and/or control and/or control rate continuity are required.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.

        Returns
        -------
        any_state_continuity : bool
            True if any state continuity is required to be enforced.
        any_control_continuity : bool
            True if any control value continuity is required to be enforced.
        any_control_rate_continuity : bool
            True if any control rate continuity is required to be enforced.
        """
        num_seg = self.grid_data.num_segments
        compressed = self.grid_data.compressed

        any_state_continuity = num_seg > 1 and not compressed
        any_control_continuity = any([opts['continuity'] for opts in phase.control_options.values()])
        any_control_continuity = any_control_continuity and num_seg > 1
        any_rate_continuity = any([opts['rate_continuity'] or opts['rate2_continuity']
                                   for opts in phase.control_options.values()])
        any_rate_continuity = any_rate_continuity and num_seg > 1

        return any_state_continuity, any_control_continuity, any_rate_continuity



import numpy as np
import openmdao.api as om


class BalancedFieldODEComp(om.ExplicitComponent):
    """
    The ODE System for an aircraft takeoff climb.

    Computes the rates for states v (true airspeed) gam (flight path angle) r (range) and h (altitude).

    References
    ----------
    .. [1] Raymer, Daniel. Aircraft design: a conceptual approach. American Institute of
    Aeronautics and Astronautics, Inc., 2012.
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('g', types=(float, int), default=9.80665, desc='gravitational acceleration (m/s**2)')
        self.options.declare('mode', values=('runway', 'climb'), desc='mode of operation (ground roll or flight)')

    def setup(self):
        nn = self.options['num_nodes']

        # Scalar (constant) inputs
        self.add_input('rho', val=1.225, desc='atmospheric density at runway', units='kg/m**3')
        self.add_input('S', val=124.7, desc='aerodynamic reference area', units='m**2')
        self.add_input('CD0', val=0.03, desc='zero-lift drag coefficient', units=None)
        self.add_input('CL0', val=0.5, desc='zero-alpha lift coefficient', units=None)
        self.add_input('CL_max', val=2.0, desc='maximum lift coefficient for linear fit', units=None)
        self.add_input('alpha_max', val=np.radians(10), desc='angle of attack at CL_max', units='rad')
        self.add_input('h_w', val=1.0, desc='height of the wing above the CG', units='m')
        self.add_input('AR', val=9.45, desc='wing aspect ratio', units=None)
        self.add_input('e', val=0.801, desc='Oswald span efficiency factor', units=None)
        self.add_input('span', val=35.7, desc='Wingspan', units='m')
        self.add_input('T', val=1.0, desc='thrust', units='N')

        # Dynamic inputs (can assume a different value at every node)
        self.add_input('m', shape=(nn,), desc='aircraft mass', units='kg')
        self.add_input('v', shape=(nn,), desc='aircraft true airspeed', units='m/s')
        self.add_input('h', shape=(nn,), desc='altitude', units='m')
        self.add_input('alpha', shape=(nn,), desc='angle of attack', units='rad')

        # Outputs
        self.add_output('CL', shape=(nn,), desc='lift coefficient', units=None)
        self.add_output('q', shape=(nn,), desc='dynamic pressure', units='Pa')
        self.add_output('L', shape=(nn,), desc='lift force', units='N')
        self.add_output('D', shape=(nn,), desc='drag force', units='N')
        self.add_output('K', val=np.ones(nn), desc='drag-due-to-lift factor', units=None)
        self.add_output('F_r', shape=(nn,), desc='runway normal force', units='N')
        self.add_output('v_dot', shape=(nn,), desc='rate of change of speed', units='m/s**2',
                        tags=['dymos.state_rate_source:v'])
        self.add_output('r_dot', shape=(nn,), desc='rate of change of range', units='m/s',
                        tags=['dymos.state_rate_source:r'])
        self.add_output('W', shape=(nn,), desc='aircraft weight', units='N')
        self.add_output('v_stall', shape=(nn,), desc='stall speed', units='m/s')
        self.add_output('v_over_v_stall', shape=(nn,), desc='stall speed ratio', units=None)

        # Mode-dependent IO
        if self.options['mode'] == 'runway':
            self.add_input('mu_r', val=0.05, desc='runway friction coefficient', units=None)
        else:
            self.add_input('gam', shape=(nn,), desc='flight path angle', units='rad')
            self.add_output('gam_dot', shape=(nn,), desc='rate of change of flight path angle',
                            units='rad/s', tags=['dymos.state_rate_source:gam'])
            self.add_output('h_dot', shape=(nn,), desc='rate of change of altitude', units='m/s',
                            tags=['dymos.state_rate_source:h'])

        self.declare_coloring(wrt='*', method='cs')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        g = self.options['g']

        # Compute factor k to include ground effect on lift
        rho = inputs['rho']
        v = inputs['v']
        S = inputs['S']
        CD0 = inputs['CD0']
        m = inputs['m']
        T = inputs['T']
        h = inputs['h']
        h_w = inputs['h_w']
        span = inputs['span']
        AR = inputs['AR']
        CL0 = inputs['CL0']
        alpha = inputs['alpha']
        alpha_max = inputs['alpha_max']
        CL_max = inputs['CL_max']
        e = inputs['e']

        outputs['W'] = W = m * g
        outputs['v_stall'] = v_stall = np.sqrt(2 * W / rho / S / CL_max)
        outputs['v_over_v_stall'] = v / v_stall

        outputs['CL'] = CL = CL0 + (alpha / alpha_max) * (CL_max - CL0)
        K_nom = 1.0 / (np.pi * AR * e)
        b = span / 2.0
        fact = ((h + h_w) / b) ** 1.5
        outputs['K'] = K = K_nom * 33 * fact / (1.0 + 33 * fact)

        outputs['q'] = q = 0.5 * rho * v ** 2
        outputs['L'] = L = q * S * CL
        outputs['D'] = D = q * S * (CD0 + K * CL ** 2)

        # Compute the downward force on the landing gear
        calpha = np.cos(alpha)
        salpha = np.sin(alpha)

        # Runway normal force
        outputs['F_r'] = F_r = m * g - L * calpha - T * salpha

        # Compute the dynamics
        if self.options['mode'] == 'climb':
            gam = inputs['gam']
            cgam = np.cos(gam)
            sgam = np.sin(gam)
            outputs['v_dot'] = (T * calpha - D) / m - g * sgam
            outputs['gam_dot'] = (T * salpha + L) / (m * v) - (g / v) * cgam
            outputs['h_dot'] = v * sgam
            outputs['r_dot'] = v * cgam
        else:
            outputs['v_dot'] = (T * calpha - D - F_r * inputs['mu_r']) / m
            outputs['r_dot'] = v



import numpy as np

from openmdao.utils.units import unit_conversion
from ...options import options as dymos_options

from ..common.timeseries_output_comp import TimeseriesOutputCompBase


class ExplicitTimeseriesComp(TimeseriesOutputCompBase):

    """
    Class definition of the ExplicitTimeseriesComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super(ExplicitTimeseriesComp, self).__init__(**kwargs)

        self.input_num_nodes = 0
        self.output_num_nodes = 0

        # Sources is used internally to map the source of a connection to the timeseries to
        # the corresponding input variable.  This is used to ensure that we don't need to connect
        # the same source to this timeseries multiple times.
        self._sources = {}

        # Used to track conversion factors for instances when one output that relies on an input
        # from another variable has potentially different units
        self._units = {}
        self._conversion_factors = {}

        self._no_check_partials = not dymos_options['include_check_partials']

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        igd = self.options['input_grid_data']
        ogd = self.options['output_grid_data']

        if ogd is not None:
            raise ValueError('Currently ExplicitTimeseriesComp does not accept a separate '
                             'GridData for output.  Leave output_grid_data as None')

        # Make the assumption here that we
        self.input_num_nodes = igd.subset_num_nodes['segment_ends']
        self.output_num_nodes = self.input_num_nodes

    def _add_output_configure(self, name, units, shape, desc='', src=None, rate=False):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        src : str
            The src path of the variables input, used to prevent redundant inputs.
        rate : bool
            If True, timeseries output is a rate.

        Returns
        -------
        bool
            True if a new input was added for the output, or False if it reuses an existing input.
        """
        if rate:
            raise NotImplementedError("Timeseries output rates are not currently supported for "
                                      "ExplicitShooting transcriptions.")

        input_num_nodes = self.input_num_nodes
        output_num_nodes = self.output_num_nodes
        added_source = False

        if name in self._vars:
            return False

        if src in self._sources:
            # If we're already pulling the source into this timeseries, use that as the
            # input for this output.
            input_name = self._sources[src]
            input_units = self._units[input_name]
        else:
            input_name = f'input_values:{name}'
            self.add_input(input_name,
                           shape=(input_num_nodes,) + shape,
                           units=units, desc=desc)
            self._sources[src] = input_name
            input_units = self._units[input_name] = units
            added_source = True

        self.add_output(name, shape=(output_num_nodes,) + shape, units=units, desc=desc)

        self._vars[name] = (input_name, name, shape, rate)

        size = np.prod(shape)
        rs = cs = np.arange(output_num_nodes * size, dtype=int)

        # There's a chance that the input for this output was pulled from another variable with
        # different units, so account for that with a conversion.
        if units is None or input_units is None:
            scale = 1.0
            offset = 0
        else:
            scale, offset = unit_conversion(input_units, units)
            self._conversion_factors[name] = scale, offset

        self.declare_partials(of=name, wrt=input_name, rows=rs, cols=cs, val=scale)

        return added_source

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for (input_name, output_name, _, _) in self._vars.values():
            if output_name in self._conversion_factors:
                scale, offset = self._conversion_factors[output_name]
                outputs[output_name] = scale * (inputs[input_name] + offset)
            else:
                outputs[output_name] = inputs[input_name]



import numpy as np

import openmdao.api as om


class CLComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('CLa', shape=(nn,), desc='alpha lift coefficient', units=None)
        self.add_input('alpha', shape=(nn,), desc='angle of attck', units='rad')

        # Outputs
        self.add_output(name='CL', val=np.ones(nn), desc='lift coefficient', units=None)

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='CL', wrt='CLa', rows=ar, cols=ar)
        self.declare_partials(of='CL', wrt='alpha', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['CL'][:] = inputs['CLa'] * inputs['alpha']

    def compute_partials(self, inputs, partials):
        partials['CL', 'CLa'] = inputs['alpha']
        partials['CL', 'alpha'] = inputs['CLa']



import os
import unittest

import openmdao.api as om

import dymos as dm

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestUpgrade_0_16_0(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_parameters(self):
        """
        # upgrade_doc: begin set_val
        p.set_val('traj.phase0.design_parameters:thrust', 2.1, units='MN')
        # upgrade_doc: end set_val
        # upgrade_doc: begin parameter_timeseries
        thrust = p.get_val('traj.phase0.timeseries.design_parameters:thrust')
        # upgrade_doc: end parameter_timeseries
        """
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Setup and solve the optimal control problem
        #
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        from dymos.examples.ssto.launch_vehicle_ode import LaunchVehicleODE

        #
        # Initialize our Trajectory and Phase
        #
        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=LaunchVehicleODE,
                         transcription=dm.GaussLobatto(num_segments=12, order=3, compressed=False))

        traj.add_phase('phase0', phase)
        p.model.add_subsystem('traj', traj)

        #
        # Set the options for the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(10, 500))

        phase.add_state('x', fix_initial=True, ref=1.0E5, defect_ref=10000.0,
                        rate_source='xdot')
        phase.add_state('y', fix_initial=True, ref=1.0E5, defect_ref=10000.0,
                        rate_source='ydot')
        phase.add_state('vx', fix_initial=True, ref=1.0E3, defect_ref=1000.0,
                        rate_source='vxdot')
        phase.add_state('vy', fix_initial=True, ref=1.0E3, defect_ref=1000.0,
                        rate_source='vydot')
        phase.add_state('m', fix_initial=True, ref=1.0E3, defect_ref=100.0,
                        rate_source='mdot')

        phase.add_control('theta', units='rad', lower=-1.57, upper=1.57)
        phase.add_parameter('thrust', units='N', opt=False, val=2100000.0)

        #
        # Set the options for our constraints and objective
        #
        phase.add_boundary_constraint('y', loc='final', equals=1.85E5, linear=True)
        phase.add_boundary_constraint('vx', loc='final', equals=7796.6961)
        phase.add_boundary_constraint('vy', loc='final', equals=0)

        phase.add_objective('time', loc='final', scaler=0.01)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup and set initial values
        #
        p.setup(check=True)

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 150.0)
        p.set_val('traj.phase0.states:x', phase.interpolate(ys=[0, 1.15E5], nodes='state_input'))
        p.set_val('traj.phase0.states:y', phase.interpolate(ys=[0, 1.85E5], nodes='state_input'))
        p.set_val('traj.phase0.states:vx', phase.interpolate(ys=[0, 7796.6961], nodes='state_input'))
        p.set_val('traj.phase0.states:vy', phase.interpolate(ys=[1.0E-6, 0], nodes='state_input'))
        p.set_val('traj.phase0.states:m', phase.interpolate(ys=[117000, 1163], nodes='state_input'))
        p.set_val('traj.phase0.controls:theta', phase.interpolate(ys=[1.5, -0.76], nodes='control_input'))

        # upgrade_doc: begin set_val
        p.set_val('traj.phase0.parameters:thrust', 2.1, units='MN')
        # upgrade_doc: end set_val

        #
        # Solve the Problem
        #
        dm.run_problem(p)

        #
        # Check the results.
        #
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 143, tolerance=0.05)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:y')[-1], 1.85E5, 1e-4)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:vx')[-1], 7796.6961, 1e-4)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:vy')[-1], 0, 1e-4)

        # upgrade_doc: begin parameter_timeseries
        thrust = p.get_val('traj.phase0.timeseries.parameters:thrust')
        # upgrade_doc: end parameter_timeseries
        nn = phase.options['transcription'].grid_data.num_nodes
        assert_near_equal(thrust, 2.1E6 * np.ones([nn, 1]))

    def test_parameter_no_timeseries(self):
        import openmdao.api as om
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('y', fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        # upgrade_doc: begin parameter_no_timeseries
        phase.add_parameter('g', opt=False, units='m/s**2', include_timeseries=False)
        # upgrade_doc: end parameter_no_timeseries

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        with self.assertRaises(KeyError):
            p.get_val('phase0.timeseries.parameters:g}')

    def test_simplified_ode_timeseries_output(self):
        """
        # upgrade_doc: begin simplified_ode_output_timeseries
        phase.add_timeseries_output('tas_comp.TAS', shape=(1,), units='m/s')
        # upgrade_doc: end simplified_ode_output_timeseries
        """
        import openmdao.api as om
        import dymos as dm
        from dymos.examples.aircraft_steady_flight.aircraft_ode import AircraftODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=1,
                                        order=13,
                                        compressed=False)
        phase = dm.Phase(ode_class=AircraftODE, transcription=transcription)
        p.model.add_subsystem('phase0', phase)

        # Pass Reference Area from an external source
        assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
        assumptions.add_output('S', val=427.8, units='m**2')
        assumptions.add_output('mass_empty', val=1.0, units='kg')
        assumptions.add_output('mass_payload', val=1.0, units='kg')

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(3600, 3600),
                               duration_ref=3600)

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(3600, 3600),
                               duration_ref=3600)

        phase.add_state('range', units='km', fix_initial=True, fix_final=False, scaler=0.01,
                        rate_source='range_rate_comp.dXdt:range',
                        defect_scaler=0.01)
        phase.add_state('mass_fuel', units='kg', fix_final=True, upper=20000.0, lower=0.0,
                        rate_source='propulsion.dXdt:mass_fuel',
                        scaler=1.0E-4, defect_scaler=1.0E-2)
        phase.add_state('alt',
                        rate_source='climb_rate',
                        units='km', fix_initial=True)

        phase.add_control('mach',  targets=['tas_comp.mach', 'aero.mach'], units=None, opt=False)

        phase.add_control('climb_rate', targets=['gam_comp.climb_rate'], units='m/s', opt=False)

        phase.add_parameter('S',
                            targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                            units='m**2')

        phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
        phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

        phase.add_path_constraint('propulsion.tau', lower=0.01, upper=1.0, shape=(1,))

        # upgrade_doc: begin simplified_ode_output_timeseries
        phase.add_timeseries_output('tas_comp.TAS')
        # upgrade_doc: end simplified_ode_output_timeseries

        p.model.connect('assumptions.S', 'phase0.parameters:S')
        p.model.connect('assumptions.mass_empty', 'phase0.parameters:mass_empty')
        p.model.connect('assumptions.mass_payload', 'phase0.parameters:mass_payload')

        phase.add_objective('time', loc='final', ref=3600)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 1.515132 * 3600.0
        p['phase0.states:range'] = phase.interpolate(ys=(0, 1296.4), nodes='state_input')
        p['phase0.states:mass_fuel'] = phase.interpolate(ys=(12236.594555, 0), nodes='state_input')
        p['phase0.states:alt'] = 5.0
        p['phase0.controls:mach'] = 0.8
        p['phase0.controls:climb_rate'] = 0.0

        p['assumptions.S'] = 427.8
        p['assumptions.mass_empty'] = 0.15E6
        p['assumptions.mass_payload'] = 84.02869 * 400

        dm.run_problem(p)

        time = p.get_val('phase0.timeseries.time')
        tas = p.get_val('phase0.timeseries.TAS', units='km/s')
        range = p.get_val('phase0.timeseries.states:range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

        exp_out = phase.simulate()

        time = exp_out.get_val('phase0.timeseries.time')
        tas = exp_out.get_val('phase0.timeseries.TAS', units='km/s')
        range = exp_out.get_val('phase0.timeseries.states:range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_glob_timeseries_outputs(self):
        """
        # upgrade_doc: begin glob_timeseries_outputs
        phase.add_timeseries_output('aero.mach', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CD0', shape=(1,), units=None)
        phase.add_timeseries_output('aero.kappa', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CLa', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CL', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CD', shape=(1,), units=None)
        phase.add_timeseries_output('aero.q', shape=(1,), units='N/m**2')
        phase.add_timeseries_output('aero.f_lift', shape=(1,), units='N')
        phase.add_timeseries_output('aero.f_drag', shape=(1,), units='N')
        # upgrade_doc: end glob_timeseries_outputs
        """
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=15, compressed=True))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                        ref=1.0E3, defect_ref=1.0E3, units='m',
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m',
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m/s',
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                        ref=1.0, defect_ref=1.0, units='rad',
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                        ref=1.0E3, defect_ref=1.0E3, units='kg',
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False, targets=['alpha'])

        phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
        phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3, units='m')
        phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0, shape=(1,))
        phase.add_boundary_constraint('gam', loc='final', equals=0.0, units='rad')

        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8, shape=(1,))

        phase.add_objective('time', loc='final', ref=1.0)

        p.model.linear_solver = om.DirectSolver()

        # upgrade_doc: begin glob_timeseries_outputs
        phase.add_timeseries_output('aero.*')
        # upgrade_doc: end glob_timeseries_outputs

        p.setup(check=True)

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 500

        p['traj.phase0.states:r'] = phase.interpolate(ys=[0.0, 50000.0], nodes='state_input')
        p['traj.phase0.states:h'] = phase.interpolate(ys=[100.0, 20000.0], nodes='state_input')
        p['traj.phase0.states:v'] = phase.interpolate(ys=[135.964, 283.159], nodes='state_input')
        p['traj.phase0.states:gam'] = phase.interpolate(ys=[0.0, 0.0], nodes='state_input')
        p['traj.phase0.states:m'] = phase.interpolate(ys=[19030.468, 10000.], nodes='state_input')
        p['traj.phase0.controls:alpha'] = phase.interpolate(ys=[0.0, 0.0], nodes='control_input')

        #
        # Solve for the optimal trajectory
        #
        p.run_model()

        outputs = p.model.list_outputs(units=True, out_stream=None, prom_name=True)
        op_dict = {options['prom_name']: options['units'] for abs_name, options in outputs}

        for name, units in [('mach', None), ('CD0', None), ('kappa', None), ('CLa', None),
                            ('CL', None), ('CD', None), ('q', 'N/m**2'), ('f_lift', 'N'),
                            ('f_drag', 'N')]:
            self.assertEqual(op_dict[f'traj.phase0.timeseries.{name}'], units)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_sequence_timeseries_outputs(self):
        """
        # upgrade_doc: begin sequence_timeseries_outputs
        phase.add_timeseries_output('aero.mach', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CD0', shape=(1,), units=None)
        phase.add_timeseries_output('aero.kappa', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CLa', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CL', shape=(1,), units=None)
        phase.add_timeseries_output('aero.CD', shape=(1,), units=None)
        phase.add_timeseries_output('aero.q', shape=(1,), units='N/m**2')
        phase.add_timeseries_output('aero.f_lift', shape=(1,), units='lbf')
        phase.add_timeseries_output('aero.f_drag', shape=(1,), units='N')
        phase.add_timeseries_output('prop.thrust', shape=(1,), units='lbf')
        # upgrade_doc: end sequence_timeseries_outputs
        # upgrade_doc: begin state_endpoint_values
        final_range = p.get_val('traj.phase0.final_conditions.states:x0++')
        final_alpha = p.get_val('traj.phase0.final_conditions.controls:alpha++')
        # upgrade_doc: end state_endpoint_values
        """
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=15, compressed=True))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                        ref=1.0E3, defect_ref=1.0E3, units='m',
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m',
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m/s',
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                        ref=1.0, defect_ref=1.0, units='rad',
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                        ref=1.0E3, defect_ref=1.0E3, units='kg',
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False, targets=['alpha'])

        phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
        phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3, units='m')
        phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0, shape=(1,))
        phase.add_boundary_constraint('gam', loc='final', equals=0.0, units='rad')

        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8, shape=(1,))

        phase.add_objective('time', loc='final', ref=1.0)

        p.model.linear_solver = om.DirectSolver()

        # upgrade_doc: begin sequence_timeseries_outputs
        phase.add_timeseries_output(['aero.*', 'prop.thrust'],
                                    units={'aero.f_lift': 'lbf', 'prop.thrust': 'lbf'})
        # upgrade_doc: end sequence_timeseries_outputs

        p.setup(check=True)

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 500

        p['traj.phase0.states:r'] = phase.interpolate(ys=[0.0, 50000.0], nodes='state_input')
        p['traj.phase0.states:h'] = phase.interpolate(ys=[100.0, 20000.0], nodes='state_input')
        p['traj.phase0.states:v'] = phase.interpolate(ys=[135.964, 283.159], nodes='state_input')
        p['traj.phase0.states:gam'] = phase.interpolate(ys=[0.0, 0.0], nodes='state_input')
        p['traj.phase0.states:m'] = phase.interpolate(ys=[19030.468, 10000.], nodes='state_input')
        p['traj.phase0.controls:alpha'] = phase.interpolate(ys=[0.0, 0.0], nodes='control_input')

        p.run_model()

        # upgrade_doc: begin state_endpoint_values
        final_range = p.get_val('traj.phase0.timeseries.states:r')[-1, ...]
        final_alpha = p.get_val('traj.phase0.timeseries.controls:alpha')[-1, ...]
        # upgrade_doc: end state_endpoint_values
        self.assertEqual(final_range, 50000.0)
        self.assertEqual(final_alpha, 0.0)

        outputs = p.model.list_outputs(units=True, out_stream=None, prom_name=True)
        op_dict = {options['prom_name']: options['units'] for abs_name, options in outputs}

        for name, units in [('mach', None), ('CD0', None), ('kappa', None), ('CLa', None),
                            ('CL', None), ('CD', None), ('q', 'N/m**2'), ('f_lift', 'lbf'),
                            ('f_drag', 'N'), ('thrust', 'lbf')]:
            self.assertEqual(op_dict[f'traj.phase0.timeseries.{name}'], units)


@use_tempdirs
class TestUpgrade_0_17_0(unittest.TestCase):

    def test_tags(self):
        """
        # upgrade_doc: begin tag_rate_source
        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s')
        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s')
        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')
        # upgrade_doc: end tag_rate_source
        # upgrade_doc: begin declare_rate_source
        phase.add_state('x', rate_source='xdot', fix_initial=True, fix_final=True)
        phase.add_state('y', rate_source='ydot', fix_initial=True, fix_final=True)
        phase.add_state('v', rate_source='vdot', fix_initial=True, fix_final=False)
        # upgrade_doc: end declare_rate_source
        """
        import numpy as np
        import openmdao.api as om

        class BrachistochroneODE(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)
                self.options.declare('static_gravity', types=(bool,), default=False,
                                     desc='If True, treat gravity as a static (scalar) input, rather than '
                                          'having different values at each node.')

            def setup(self):
                nn = self.options['num_nodes']
                g_default_val = 9.80665 if self.options['static_gravity'] else 9.80665 * np.ones(nn)

                # Inputs
                self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

                self.add_input('g', val=g_default_val, desc='grav. acceleration', units='m/s/s')

                self.add_input('theta', val=np.ones(nn), desc='angle of wire', units='rad')

                # upgrade_doc: begin tag_rate_source
                self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s',
                                tags=['dymos.state_rate_source:x', 'dymos.state_units:m'])

                self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s',
                                tags=['dymos.state_rate_source:y', 'dymos.state_units:m'])

                self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2',
                                tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])
                # upgrade_doc: end tag_rate_source

                self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                                units='m/s')

                # Setup partials
                arange = np.arange(self.options['num_nodes'])
                self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

                self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
                self.declare_partials(of='xdot', wrt='theta', rows=arange, cols=arange)

                self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
                self.declare_partials(of='ydot', wrt='theta', rows=arange, cols=arange)

                self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
                self.declare_partials(of='check', wrt='theta', rows=arange, cols=arange)

                if self.options['static_gravity']:
                    c = np.zeros(self.options['num_nodes'])
                    self.declare_partials(of='vdot', wrt='g', rows=arange, cols=c)
                else:
                    self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)

            def compute(self, inputs, outputs):
                theta = inputs['theta']
                cos_theta = np.cos(theta)
                sin_theta = np.sin(theta)
                g = inputs['g']
                v = inputs['v']

                outputs['vdot'] = g * cos_theta
                outputs['xdot'] = v * sin_theta
                outputs['ydot'] = -v * cos_theta
                outputs['check'] = v / sin_theta

            def compute_partials(self, inputs, partials):
                theta = inputs['theta']
                cos_theta = np.cos(theta)
                sin_theta = np.sin(theta)
                g = inputs['g']
                v = inputs['v']

                partials['vdot', 'g'] = cos_theta
                partials['vdot', 'theta'] = -g * sin_theta

                partials['xdot', 'v'] = sin_theta
                partials['xdot', 'theta'] = v * cos_theta

                partials['ydot', 'v'] = -cos_theta
                partials['ydot', 'theta'] = v * sin_theta

                partials['check', 'v'] = 1 / sin_theta
                partials['check', 'theta'] = -v * cos_theta / sin_theta ** 2

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        #
        # Initialize the Problem and the optimization driver
        #
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        transcription=dm.GaussLobatto(num_segments=10)))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        # upgrade_doc: begin declare_rate_source
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)
        # upgrade_doc: end declare_rate_source

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)
        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        #
        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 2.0

        p['traj.phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['traj.phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['traj.phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['traj.phase0.controls:theta'] = phase.interpolate(ys=[5, 100.5], nodes='control_input')

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p)

        # Test the results
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)


@use_tempdirs
class TestUpgrade_0_19_0(unittest.TestCase):

    def tearDown(self):
        if os.path.exists('dymos_solution.db'):
            os.remove('dymos_solution.db')
        if os.path.exists('dymos_simulation.db'):
            os.remove('dymos_simulation.db')

    @require_pyoptsparse(optimizer='SLSQP')
    def _make_problem(self, transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                      compressed=True, optimizer='SLSQP', run_driver=True,
                      force_alloc_complex=False,
                      solve_segments=False):

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring(tol=1.0E-12)

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)

        # upgrade_doc: begin exec_comp_ode
        ode = lambda num_nodes: om.ExecComp(['vdot = g * cos(theta)',
                                             'xdot = v * sin(theta)',
                                             'ydot = -v * cos(theta)'],
                                            g={'val': 9.80665, 'units': 'm/s**2'},
                                            v={'shape': (num_nodes,), 'units': 'm/s'},
                                            theta={'shape': (num_nodes,), 'units': 'rad'},
                                            vdot={'shape': (num_nodes,), 'units': 'm/s**2'},
                                            xdot={'shape': (num_nodes,), 'units': 'm/s'},
                                            ydot={'shape': (num_nodes,), 'units': 'm/s'},
                                            has_diag_partials=True)

        phase = dm.Phase(ode_class=ode, transcription=t)
        # upgrade_doc: end declare_rate_source
        traj = dm.Trajectory()
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='ydot')

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='vdot')

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', static_target=True)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.setup(check=['unconnected_inputs'], force_alloc_complex=force_alloc_complex)

        p['traj0.phase0.t_initial'] = 0.0
        p['traj0.phase0.t_duration'] = 2.0

        p['traj0.phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['traj0.phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['traj0.phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['traj0.phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['traj0.phase0.parameters:g'] = 9.80665

        dm.run_problem(p, run_driver=run_driver, simulate=True)

        return p

    def run_asserts(self):

        for db in ['dymos_solution.db', 'dymos_simulation.db']:
            p = om.CaseReader(db).get_case('final')

            t_initial = p.get_val('traj0.phase0.timeseries.time')[0]
            tf = p.get_val('traj0.phase0.timeseries.time')[-1]

            x0 = p.get_val('traj0.phase0.timeseries.states:x')[0]
            xf = p.get_val('traj0.phase0.timeseries.states:x')[-1]

            y0 = p.get_val('traj0.phase0.timeseries.states:y')[0]
            yf = p.get_val('traj0.phase0.timeseries.states:y')[-1]

            v0 = p.get_val('traj0.phase0.timeseries.states:v')[0]
            vf = p.get_val('traj0.phase0.timeseries.states:v')[-1]

            g = p.get_val('traj0.phase0.timeseries.parameters:g')[0]

            thetaf = p.get_val('traj0.phase0.timeseries.controls:theta')[-1]

            assert_near_equal(t_initial, 0.0)
            assert_near_equal(x0, 0.0)
            assert_near_equal(y0, 10.0)
            assert_near_equal(v0, 0.0)

            assert_near_equal(tf, 1.8016, tolerance=0.01)
            assert_near_equal(xf, 10.0, tolerance=0.01)
            assert_near_equal(yf, 5.0, tolerance=0.01)
            assert_near_equal(vf, 9.902, tolerance=0.01)
            assert_near_equal(g, 9.80665, tolerance=0.01)

            assert_near_equal(thetaf, 100.12, tolerance=0.01)

    def test_ex_brachistochrone_radau_uncompressed(self):
        self._make_problem(transcription='radau-ps', compressed=False)
        self.run_asserts()

    def test_ex_brachistochrone_gl_uncompressed(self):
        self._make_problem(transcription='gauss-lobatto', compressed=False)
        self.run_asserts()
